function x = US_Rec(modulo_data, data, L, DO);


% US_Rec: Unlimited Sampling Reconstruction Algorithm
% From the paper:

% 2017: On Unlimited Sampling (DOI: 10.1109/SAMPTA.2017.8024471)
% 2020: On Unlimited Sampling and Reconstruction (DOI:
% 10.1109/TSP.2020.3041955)



% Inputs:
%   modulo_data - Modulo sampled data.
%   data        - Original signal data for reference.
%   L           - Modulo threshold parameter.
%   DO          - Difference order.
%
% Output:
%   x           - Reconstructed signal.

% Euler's number definition
 e = exp(1);
% Define fractional part function
frac = @(x)  x - floor(x);
% Origin-centered rounding function
RD = @(x,L) round(x ./ (2*L)) .* 2*L;
% General modulo function
M = @(f,T) 2*T .* (frac(f ./ (2 .* T) + 0.5) - 0.5);

% Compute differences of modulo data
 ds = diff(modulo_data, DO);
 rec = M(ds, L);
 res = rec(:) - ds(:);
 res = RD(res, L);

% Real data-based centering
 z0 = (res - RD(res, L));

% Initialize counter
 cntr = cell(DO);
 
 % Clear intermediate variables
 clear S2 Kn_iter Rec_Sig al bl egN cntr

% Initialize variables for iterative reconstruction
 al = res;                    
 bl = res;                    
 egN = al;                   

% Calculate bound for iterations
 g0 = ceil(max(abs(data)));
 Bg = ceil(g0 / (2*L)) * 2*L;
 J = ceil(6 * Bg / L);

% Backward reconstruction loop
for k = 1:DO-1
    cntr{k} = egN;                                  % Store iteration data
    S2 = cumsum([0; (cumsum([0; egN]))]);           % Double cumulative sum
    Kn_iter = floor((-S2(J+1) + S2(1)) / (12*Bg) + 0.5);
    bl = cumsum([0; bl]);                           % Cumulative sum of reconstruction
    bl = RD(bl, L);                                 % Round to nearest modulo range
    bl = bl + 2 * Kn_iter * L;                      % Adjust based on iteration counter
    egN = bl;                                       % Update egN for next iteration
end

% Final signal reconstruction and offset adjustment
Rec_Sig = cumsum([0; bl]) + modulo_data(:);
offset = mean(data - Rec_Sig);
Rec_Sig = Rec_Sig + offset;

% Output reconstructed signal
x = Rec_Sig;
